/* Make it possible to have multiline table cells by removing white-space:nowrap */
.wy-table-responsive table td, .wy-table-responsive table th {
    white-space: normal;
}
.wy-table-responsive {
    margin-bottom: 24px;
    max-width: 100%;
    overflow: visible;
}
